#!/usr/bin/python
"""Classes to handle sodar data samples

Sodar data samples are collected into daily files. Each sample consists of a
header followed by an observation for each height.
"""

import re

class Data(list):
    """Daily sodar file data

       (a collection of samples)
    """
    def __init__(self,data):
        super(Data,self).__init__()
        # samples in file are terminated by $
        self.extend([Sample(sample)
                     for sample in
                     [sample.strip() for sample in data.split('$')]
                     if sample])

    def __getitem__(self,index):
        """allow sample retrieval by sample time in header
        """
        try:
            return super(Data,self).__getitem__(index)
        except TypeError:
            return self.find(index)

    def find(self,index):
        """find Sample in Data

            where sample time of form YYYY-MM-DD-HH-MM
        """
        try:
            year,month,day,hour,min = index.split('-')
        except ValueError:
            raise ValueError,'Data index by date must be "YYYY-MM-DD-HH-MM"'
        except AttributeError:
            raise AttributeError,'Data index by date must be "YYYY-MM-DD-HH-MM"'
        for sample in self:
            if sample.header['YEAR'].rjust(4,'0') != year: continue
            if sample.header['MONTH'].rjust(2,'0') != month: continue
            if sample.header['DAY'].rjust(2,'0') != day: continue
            if sample.header['HOUR'].rjust(2,'0') != hour: continue
            if sample.header['MIN'].rjust(2,'0') != min: continue
            return sample
        raise IndexError,'Data index out of range'

class Sample(object):
    """A single sample from daily sodar file data

       (a header and a body)
    """
    def __init__(self,sample):
        super(Sample,self).__init__()
        # first three groups of lines are the header; rest is body
        samplePattern = re.compile(r'''(?P<header>.*?\n\n.*?\n\n.*?\n\n)
                                       (?P<body>.*$)
                                    ''',re.DOTALL | re.VERBOSE)
        self.__dict__.update(samplePattern.match(sample.strip()).groupdict())
        # fix for missing keys
        self.header = Header(self.header)
        self.body = Body(self.body)

    def __getitem__(self,index):
        return self.body[index]

class Header(dict):
    """A sodar data sample header

      (a collection of sample-wide parameters)
    """
    def __init__(self,header):
        super(Header,self).__init__()
        headerLines = [headerLine.strip()
                       for headerLine in header.split('\n')
                       if headerLine]
        # every other line contains parameter names;
        # every other line contains parameter values
        parametersPairs = [(headerLine,
                            headerLines[headerLines.index(headerLine)+1])
                           for headerLine in headerLines[::2]]
        for parameterNames,parameterValues in parametersPairs:
            # parameter names must be valid Python identifiers
            # for named groups matching
            parameterNames = [parameterName.strip('#')
                              for parameterName in parameterNames.split()]
            parameterPattern = re.compile(r'(?P<' +
                                          '>\S+)\s+(?P<'.join(parameterNames) +
                                          '>.*$)')
            self.update(parameterPattern.
                        match(parameterValues).
                        groupdict())

class Body(list):
    """A sodar data sample body

       (a collection of collections at each altitude)
    """
    def __init__(self,body):
        super(Body,self).__init__()
        bodyLines = [bodyLine.strip()
                     for bodyLine in body.split('\n')
                     if bodyLine]
        parameterNames = [parameterName
                          for parameterName in bodyLines[0].split()]
        parameterPattern = re.compile(r'(?P<' +
                                      '>\S+)\s+(?P<'.join(parameterNames) +
                                      '>.*$)')
        self.extend([parameterPattern.match(bodyLine).groupdict()
                     for bodyLine in bodyLines[1:]])
        self.reverse()            

    def __getitem__(self,index):
        """allow retrieval by altitude string
        """
        try:
            return super(Body,self).__getitem__(index)
        except TypeError:
            return self.find(index)

    def find(self,index):
        """find altitude data in Body
        """
        for altitudeData in self:
            if altitudeData['ALT'] != index: continue
            return altitudeData
        raise IndexError,'Body index, out of range'

def __main():
    """Process as script from command line
    """
    import urllib2
    try:
       data = urllib2.urlopen('http://nemo.isis.unc.edu/data/nccoos/level0/dukeforest/sodar/store/2007-06/20070601.dat')
       data = data.read()
    except:
        print "Failure to read test data"
    data = Data(data)
    print '   --- Sample ---\n'.join(data.samples)

if __name__ == "__main__":
    __main()
